@@ -31,10 +31,15 @@ import usersdialog
 import util
 import userdb
 import protosession
+import debuglog
+import errors
 from config import *
 
 def dprint (fmt, *args):
-    util.debug_print (util.DEBUG_ADMINTOOL, fmt % args)
+    debuglog.debug_log (False, debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL, fmt % args)
+
+def mprint (fmt, *args):
+    debuglog.debug_log (True, debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL, fmt % args)
 
 def _get_profile_path_for_name (profile_name):
     return os.path.join (PROFILESDIR, profile_name + ".zip")
@@ -43,10 +48,10 @@ class Session (gobject.GObject):
     __gsignals__ = {
         "finished" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
         }
-        
+
     def __init__ (self, username, profile_name):
         gobject.GObject.__init__ (self)
-        
+
         self.username     = username
         self.profile_name = profile_name
         self.profile_path = _get_profile_path_for_name (profile_name)
@@ -56,8 +61,11 @@ class Session (gobject.GObject):
         self.user_profile_path = None
         self.temp_homedir      = None
 
-        self.session_pid         = 0
-        self.session_child_watch = 0
+        self.session_pid             = 0
+        self.session_child_watch     = 0
+        self.session_stderr          = None
+        self.session_stderr_watch_id = 0
+        self.session_log_str         = ""
 
     def __del__ (self):
         if self.temp_homedir:
@@ -80,7 +88,7 @@ class Session (gobject.GObject):
         os.close (fd)
 
         shutil.copy2 (xauthority, temp_path)
-        
+
         os.chown (temp_path, self.pw.pw_uid, self.pw.pw_gid)
 
         dprint ("Copied $XAUTHORITY (%s) temporarily to %s", xauthority, temp_path)
@@ -92,7 +100,7 @@ class Session (gobject.GObject):
         os.close (fd)
 
         shutil.copy2 (profile_path, user_path)
-        
+
         os.chown (user_path, self.pw.pw_uid, self.pw.pw_gid)
 
         dprint ("Copied %s temporarily to %s", profile_path, user_path)
@@ -104,72 +112,134 @@ class Session (gobject.GObject):
         shutil.move (user_path, profile_path)
         dprint ("Moved %s back from %s", user_path, profile_path)
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL)
     def __session_child_watch_handler (self, pid, status):
-        dprint ("sabayon-session died")
+        if not os.WIFEXITED (status):
+            exit_code = util.EXIT_CODE_FATAL
 
+            if os.WIFSIGNALED (status):
+                signal_num = os.WTERMSIG (status)
+
+                mprint ("sabayon-session exited with SIGNAL %s", signal_num)
+            else:
+                mprint ("sabayon-session exited for an unknown reason")
+        else:
+            exit_code = os.WEXITSTATUS (status)
+
+            # Here we cannot throw an exception for the error cases,
+            # since we are a callback running out of the Glib main
+            # loop.  So we have to set a flag and let the code
+            # *outside* the main loop pick it up.
+
+            if exit_code == util.EXIT_CODE_NORMAL:
+                mprint ("sabayon-session exited normally")
+                success = True
+            elif exit_code == util.EXIT_CODE_RECOVERABLE:
+                errors.errors_log_recoverable_error (debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL,
+                                                     "sabayon-session exited with RECOVERABLE exit status")
+                # FIXME: throw a warning dialog
+            else:
+                errors.errors_log_fatal_error (debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL,
+                                               "sabayon-session exited with a FATAL ERROR (exit code %s)" % exit_code)
+                gtk.main_quit () # so that the toplevel 'sabayon' will exit the main loop and show the fatal error
+
+        protosession.clobber_user_processes (self.username)
         protosession.reset_shell_and_homedir (self.username, self.temp_homedir)
         self.temp_homedir = None
 
         if self.temp_xauth_path:
             os.remove (self.temp_xauth_path)
             self.temp_xauth_path = None
-        
+
         self.__copy_from_user (self.user_profile_path, self.profile_path)
         self.user_profile_path = None
 
         self.session_pid         = 0
         self.session_child_watch = 0
 
+        gobject.source_remove (self.session_stderr_watch_id)
+        self.session_stderr_watch_id = 0
+
+        self.session_stderr.close ()
+        self.session_stderr = None
+
         self.emit ("finished")
-        
+
         return False
 
+    def build_envp_for_child (self):
+        # gobject.spawn_async() wants a sequence, not a dictionary, so we build a sequence...
+        new_environ = []
+        for key in PASSTHROUGH_ENVIRONMENT:
+            if os.environ.has_key (key):
+                new_environ.append ("%s=%s" % (key, os.environ[key]))
+
+        new_environ = new_environ + ["PATH=%s"     % DEFAULT_PATH,
+                                     "SHELL=%s"    % DEFAULT_SHELL,
+                                     "DISPLAY=%s"  % os.environ["DISPLAY"],
+                                     "HOME=%s"     % self.temp_homedir,
+                                     "LOGNAME=%s"  % self.pw.pw_name,
+                                     "USER=%s"     % self.pw.pw_name,
+                                     "USERNAME=%s" % self.pw.pw_name]
+
+        if self.temp_xauth_path:
+            new_environ.append ("XAUTHORITY=%s" % self.temp_xauth_path)
+
+        return new_environ
+
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL)
+    def session_stderr_io_cb (self, source_fd, condition, session):
+        if condition & gobject.IO_IN:
+            s = session.session_stderr.read ()
+            session.session_log_str = session.session_log_str + s
+#            print "%s: got from sabayon-session stderr: \n<BEGIN SABAYON-SESSION STDERR>\n%s\n<END SABAYON-SESSION STDERR>" % (os.getpid (), s)
+
+        if condition & gobject.IO_HUP:
+            mprint ("========== BEGIN SABAYON-SESSION LOG ==========\n"
+                    "%s\n"
+                    "========== END SABAYON-SESSION LOG ==========",
+                    session.session_log_str)
+            return False
+
+        return True
+
     def start (self):
         self.user_profile_path = self.__copy_to_user (self.profile_path)
         self.temp_homedir = protosession.setup_shell_and_homedir (self.username)
         protosession.clobber_user_processes (self.username)
-        
+
         display_number = protosession.find_free_display ()
 
         self.temp_xauth_path = self.__copy_xauthority ()
 
-        self.session_pid = os.fork ()
-        if self.session_pid == 0: # Child process
+        def child_setup_fn (self):
             os.setgid (self.pw.pw_gid)
             os.setuid (self.pw.pw_uid)
-
-            os.chdir (self.temp_homedir)
-
             os.setsid ()
             os.umask (022)
 
-            new_environ = {}
-            for key in PASSTHROUGH_ENVIRONMENT:
-                if os.environ.has_key (key):
-                    new_environ[key] = os.environ[key]
+        # FIXME: get_readable_log_config_filename() doesn't work here.
+        # Create a temporary copy of the log config file and use *that*.
+        argv = SESSION_TOOL_ARGV + [ ("--admin-log-config=%s" % util.get_admin_log_config_filename ()),
+                                     ("--readable-log-config=%s" % util.get_readable_log_config_filename ()),
+                                     self.profile_name,
+                                     self.user_profile_path,
+                                     str (display_number) ]
+        envp = self.build_envp_for_child ()
+        cwd = self.temp_homedir
 
-            new_environ["PATH"]       = DEFAULT_PATH
-            new_environ["SHELL"]      = DEFAULT_SHELL
-            new_environ["DISPLAY"]    = os.environ["DISPLAY"]
-            new_environ["HOME"]       = self.temp_homedir
-            new_environ["LOGNAME"]    = self.pw.pw_name
-            new_environ["USER"]       = self.pw.pw_name
-            new_environ["USERNAME"]   = self.pw.pw_name
-
-            if self.temp_xauth_path:
-                new_environ["XAUTHORITY"] = self.temp_xauth_path
-
-            dprint ("Child process env: %s" % new_environ)
-
-            argv = SESSION_TOOL_ARGV + [ self.profile_name, self.user_profile_path, str (display_number) ]
-
-            dprint ("Executing %s" % argv)
-            os.execve (SESSION_TOOL_ARGV[0], argv, new_environ)
-
-            # Shouldn't ever happen
-            sys.stderr.write ("Failed to launch sabayon-session")
-            os._exit (1)
+        # FIXME: do we need any special processing if this throws an exception?
+        # We'll catch it in the toplevel and exit with a fatal error code, anyway.
+        (pid, oink, oink, stderr_fd) = gobject.spawn_async (argv, envp, cwd,
+                                                            gobject.SPAWN_DO_NOT_REAP_CHILD,
+                                                            child_setup_fn, self,
+                                                            None, None, True)	# stdin, stdout, stderr
 
+        self.session_pid = pid;
+        self.session_stderr = os.fdopen (stderr_fd)
+        self.session_stderr_watch_id = gobject.io_add_watch (stderr_fd,
+                                                             gobject.IO_IN | gobject.IO_HUP,
+                                                             self.session_stderr_io_cb, self)
         self.session_child_watch = gobject.child_watch_add (self.session_pid,
                                                             self.__session_child_watch_handler)
 
@@ -195,10 +265,10 @@ class ProfilesModel (gtk.ListStore):
 class AddProfileDialog:
     def __init__ (self, profiles_model):
         self.profiles_model = profiles_model
-        
+
         glade_file = os.path.join (GLADEDIR, "sabayon.glade")
         self.xml = gtk.glade.XML (glade_file, "add_profile_dialog")
-        
+
         self.dialog = self.xml.get_widget ("add_profile_dialog")
         self.dialog.connect ("destroy", gtk.main_quit)
         self.dialog.set_default_response (gtk.RESPONSE_ACCEPT)
@@ -210,7 +280,7 @@ class AddProfileDialog:
         self.name_entry = self.xml.get_widget ("add_profile_name_entry")
         self.name_entry.connect ("changed", self.__name_entry_changed)
         self.name_entry.set_activates_default (True)
-        
+
         self.base_combo = self.xml.get_widget ("add_profile_base_combo")
         self.base_combo.set_model (self.profiles_model)
         if self.profiles_model.get_iter_first () is None:
@@ -220,6 +290,7 @@ class AddProfileDialog:
         self.base_combo.pack_start (renderer, True)
         self.base_combo.set_attributes (renderer, text = ProfilesModel.COLUMN_NAME)
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __name_entry_changed (self, entry):
         text = entry.get_text ()
         if not text or text.isspace ():
@@ -233,7 +304,7 @@ class AddProfileDialog:
         self.dialog.present ()
         response = self.dialog.run ()
         self.dialog.hide ()
-        
+
         if response != gtk.RESPONSE_ACCEPT:
             return (None, None)
 
@@ -242,13 +313,13 @@ class AddProfileDialog:
             base = self.profiles_model.get_value (row, ProfilesModel.COLUMN_NAME)
         else:
             base = None
-        
+
         return (self.name_entry.get_text (), base)
 
 class ProfilesDialog:
     def __init__ (self):
         assert os.geteuid () == 0
-        
+
         glade_file = os.path.join (GLADEDIR, "sabayon.glade")
         self.xml = gtk.glade.XML (glade_file, "profiles_dialog")
 
@@ -259,7 +330,7 @@ class ProfilesDialog:
 
         self.profiles_list = self.xml.get_widget ("profiles_list")
         self.__setup_profiles_list ()
-        
+
         self.profiles_list.connect ("key-press-event", self.__handle_key_press)
 
         self.add_button = self.xml.get_widget ("add_button")
@@ -267,19 +338,19 @@ class ProfilesDialog:
 
         self.remove_button = self.xml.get_widget ("remove_button")
         self.remove_button.connect ("clicked", self.__remove_button_clicked)
-        
+
         self.edit_button = self.xml.get_widget ("edit_button")
         self.__fix_button_align (self.edit_button)
         self.edit_button.connect ("clicked", self.__edit_button_clicked)
-        
+
         self.details_button = self.xml.get_widget ("details_button")
         self.__fix_button_align (self.details_button)
         self.details_button.connect ("clicked", self.__details_button_clicked)
-        
+
         self.users_button = self.xml.get_widget ("users_button")
         self.__fix_button_align (self.users_button)
         self.users_button.connect ("clicked", self.__users_button_clicked)
-        
+
         self.help_button = self.xml.get_widget ("help_button")
         self.help_button.hide ()
 
@@ -315,9 +386,11 @@ class ProfilesDialog:
                                 text = ProfilesModel.COLUMN_NAME)
         self.profiles_list.append_column (c)
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __dialog_response (self, dialog, response_id):
         dialog.destroy ()
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __add_button_clicked (self, button):
         (profile_name, base_profile) = AddProfileDialog (self.profiles_model).run (self.dialog)
         if profile_name:
@@ -329,9 +402,12 @@ class ProfilesDialog:
             return None
         return model[row][ProfilesModel.COLUMN_NAME]
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_ADMIN_TOOL)
     def __session_finished (self, session):
+        debuglog.uprint ("Finishing editing profile")
         self.dialog.set_sensitive (True)
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __edit_button_clicked (self, button):
         profile_name = self.__get_selected_profile ()
         if profile_name:
@@ -339,13 +415,16 @@ class ProfilesDialog:
 
             session = Session (PROTOTYPE_USER, profile_name)
             session.connect ("finished", self.__session_finished)
+            debuglog.uprint ("Starting to edit profile '%s'", profile_name)
             session.start ()
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __details_button_clicked (self, button):
         profile_name = self.__get_selected_profile ()
         if profile_name:
             editorwindow.ProfileEditorWindow (profile_name, self.dialog)
-        
+
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __users_button_clicked (self, button):
         profile_name = self.__get_selected_profile ()
         if profile_name:
@@ -376,7 +455,7 @@ class ProfilesDialog:
             for user in db.get_users ():
                 if db.get_profile (user, False, True) == profile_name:
                     db.set_profile (user, None)
-            
+
             self.profiles_model.reload ()
 
             row = None
@@ -391,9 +470,11 @@ class ProfilesDialog:
             if row:
                 self.profiles_list.get_selection ().select_iter (row)
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __remove_button_clicked (self, button):
         self.__delete_currently_selected ()
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __handle_key_press (self, profiles_list, event):
         if event.keyval in (gtk.keysyms.Delete, gtk.keysyms.KP_Delete):
             self.__delete_currently_selected ()
@@ -412,12 +493,12 @@ class ProfilesDialog:
             #
             name = _("%s (%s)") % (profile_name, idx)
             idx += 1
-        
+
         return name
 
     def __create_new_profile (self, profile_name, base_profile):
         profile_name = self.__make_unique_profile_name (profile_name)
-        
+
         if base_profile:
             base_storage = storage.ProfileStorage (base_profile)
             new_storage = base_storage.copy (profile_name)
@@ -433,6 +514,7 @@ class ProfilesDialog:
                 return
             row = self.profiles_model.iter_next (row)
 
+    @errors.checked_callback (debuglog.DEBUG_LOG_DOMAIN_USER)
     def __profile_selection_changed (self, selection):
         profile_name = self.__get_selected_profile ()
         self.edit_button.set_sensitive (profile_name != None)
